from __future__ import annotations

import typing as t

from piccolo.columns.column_types import ForeignKey
from piccolo.columns.combination import And, Where
from piccolo.custom_types import Combinable, TableInstance
from piccolo.engine.base import Batch
from piccolo.query.base import Query
from piccolo.query.methods.select import Select
from piccolo.query.mixins import (
    AsOfDelegate,
    CallbackDelegate,
    CallbackType,
    LimitDelegate,
    OffsetDelegate,
    OrderByDelegate,
    OrderByRaw,
    OutputDelegate,
    PrefetchDelegate,
    WhereDelegate,
)
from piccolo.query.proxy import Proxy
from piccolo.querystring import QueryString
from piccolo.utils.dictionary import make_nested
from piccolo.utils.sync import run_sync

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.columns import Column


###############################################################################


class GetOrCreate(
    Proxy["Objects[TableInstance]", TableInstance], t.Generic[TableInstance]
):
    def __init__(
        self,
        query: Objects[TableInstance],
        table_class: t.Type[TableInstance],
        where: Combinable,
        defaults: t.Dict[Column, t.Any],
    ):
        self.query = query
        self.table_class = table_class
        self.where = where
        self.defaults = defaults

    async def run(
        self, node: t.Optional[str] = None, in_pool: bool = True
    ) -> TableInstance:
        """
        :raises ValueError:
            If more than one matching row is found.

        """
        instance = await self.query.get(self.where).run(
            node=node, in_pool=in_pool
        )
        if instance:
            instance._was_created = False
            return instance

        instance = self.table_class(_data=self.defaults)

        # If it's a complex `where`, there can be several column values to
        # extract e.g. (Band.name == 'Pythonistas') & (Band.popularity == 1000)
        if isinstance(self.where, Where):
            setattr(
                instance,
                self.where.column._meta.name,  # type: ignore
                self.where.value,  # type: ignore
            )
        elif isinstance(self.where, And):
            for column, value in self.where.get_column_values().items():
                if len(column._meta.call_chain) == 0:
                    # Make sure we only set the value if the column belongs
                    # to this table.
                    setattr(instance, column._meta.name, value)

        await instance.save().run(node=node, in_pool=in_pool)

        # If the user wants us to prefetch related objects, for example:
        #
        # await Band.objects(Band.manager).get_or_create(
        #   (Band.name == 'Pythonistas') & (Band.manager == 1)
        # )
        #
        # Then we need to fetch the related objects.
        # See https://github.com/piccolo-orm/piccolo/issues/597
        prefetch = self.query.prefetch_delegate.fk_columns
        if prefetch:
            table = instance.__class__
            primary_key = table._meta.primary_key
            instance = (
                await table.objects(*prefetch)
                .get(primary_key == getattr(instance, primary_key._meta.name))
                .run()
            )

        instance = t.cast(TableInstance, instance)
        instance._was_created = True
        return instance


class Get(
    Proxy["First[TableInstance]", t.Optional[TableInstance]],
    t.Generic[TableInstance],
):
    pass


class First(
    Proxy["Objects[TableInstance]", t.Optional[TableInstance]],
    t.Generic[TableInstance],
):
    async def run(
        self, node: t.Optional[str] = None, in_pool: bool = True
    ) -> t.Optional[TableInstance]:
        objects = await self.query.run(
            node=node, in_pool=in_pool, use_callbacks=False
        )

        results = objects[0] if objects else None

        modified_response: t.Optional[
            TableInstance
        ] = await self.query.callback_delegate.invoke(
            results=results, kind=CallbackType.success
        )
        return modified_response


class Create(t.Generic[TableInstance]):
    """
    This is provided as a simple convenience. Rather than running::

        band = Band(name='Pythonistas')
        await band.save()

    We can instead do it in a single line::

        band = Band.objects().create(name='Pythonistas')

    """

    def __init__(
        self,
        table_class: t.Type[TableInstance],
        columns: t.Dict[str, t.Any],
    ):
        self.table_class = table_class
        self.columns = columns

    async def run(
        self,
        node: t.Optional[str] = None,
        in_pool: bool = True,
    ) -> TableInstance:
        instance = self.table_class(**self.columns)
        await instance.save().run(node=node, in_pool=in_pool)
        return instance

    def __await__(self) -> t.Generator[None, None, TableInstance]:
        """
        If the user doesn't explicity call .run(), proxy to it as a
        convenience.
        """
        return self.run().__await__()

    def run_sync(self, *args, **kwargs) -> TableInstance:
        return run_sync(self.run(*args, **kwargs))


###############################################################################


class Objects(
    Query[TableInstance, t.List[TableInstance]], t.Generic[TableInstance]
):
    """
    Almost identical to select, except you have to select all fields, and
    table instances are returned, rather than just data.
    """

    __slots__ = (
        "nested",
        "as_of_delegate",
        "limit_delegate",
        "offset_delegate",
        "order_by_delegate",
        "output_delegate",
        "callback_delegate",
        "prefetch_delegate",
        "where_delegate",
    )

    def __init__(
        self,
        table: t.Type[TableInstance],
        prefetch: t.Sequence[t.Union[ForeignKey, t.List[ForeignKey]]] = (),
        **kwargs,
    ):
        super().__init__(table, **kwargs)
        self.as_of_delegate = AsOfDelegate()
        self.limit_delegate = LimitDelegate()
        self.offset_delegate = OffsetDelegate()
        self.order_by_delegate = OrderByDelegate()
        self.output_delegate = OutputDelegate()
        self.output_delegate._output.as_objects = True
        self.callback_delegate = CallbackDelegate()
        self.prefetch_delegate = PrefetchDelegate()
        self.prefetch(*prefetch)
        self.where_delegate = WhereDelegate()

    def output(self: Self, load_json: bool = False) -> Self:
        self.output_delegate.output(
            as_list=False, as_json=False, load_json=load_json
        )
        return self

    def callback(
        self: Self,
        callbacks: t.Union[t.Callable, t.List[t.Callable]],
        *,
        on: CallbackType = CallbackType.success,
    ) -> Self:
        self.callback_delegate.callback(callbacks, on=on)
        return self

    def as_of(self, interval: str = "-1s") -> Objects:
        self.as_of_delegate.as_of(interval)
        return self

    def limit(self: Self, number: int) -> Self:
        self.limit_delegate.limit(number)
        return self

    def prefetch(
        self: Self, *fk_columns: t.Union[ForeignKey, t.List[ForeignKey]]
    ) -> Self:
        self.prefetch_delegate.prefetch(*fk_columns)
        return self

    def offset(self: Self, number: int) -> Self:
        self.offset_delegate.offset(number)
        return self

    def order_by(
        self: Self, *columns: t.Union[Column, str, OrderByRaw], ascending=True
    ) -> Self:
        _columns: t.List[t.Union[Column, OrderByRaw]] = []
        for column in columns:
            if isinstance(column, str):
                _columns.append(self.table._meta.get_column_by_name(column))
            else:
                _columns.append(column)

        self.order_by_delegate.order_by(*_columns, ascending=ascending)
        return self

    def where(self: Self, *where: Combinable) -> Self:
        self.where_delegate.where(*where)
        return self

    ###########################################################################

    def first(self: Self) -> First[TableInstance]:
        self.limit_delegate.limit(1)
        return First[TableInstance](query=self)

    def get(self: Self, where: Combinable) -> Get[TableInstance]:
        self.where_delegate.where(where)
        self.limit_delegate.limit(1)
        return Get[TableInstance](query=First[TableInstance](query=self))

    def get_or_create(
        self: Self,
        where: Combinable,
        defaults: t.Dict[Column, t.Any] = None,
    ) -> GetOrCreate[TableInstance]:
        if defaults is None:
            defaults = {}
        return GetOrCreate[TableInstance](
            query=self, table_class=self.table, where=where, defaults=defaults
        )

    def create(self: Self, **columns: t.Any) -> Create[TableInstance]:
        return Create[TableInstance](table_class=self.table, columns=columns)

    ###########################################################################

    async def batch(
        self: Self,
        batch_size: t.Optional[int] = None,
        node: t.Optional[str] = None,
        **kwargs,
    ) -> Batch:
        if batch_size:
            kwargs.update(batch_size=batch_size)
        if node:
            kwargs.update(node=node)
        return await self.table._meta.db.batch(self, **kwargs)

    async def response_handler(self, response):
        if self.output_delegate._output.nested:
            return [make_nested(i) for i in response]
        else:
            return response

    @property
    def default_querystrings(self) -> t.Sequence[QueryString]:
        select = Select(table=self.table)

        for attr in (
            "as_of_delegate",
            "limit_delegate",
            "where_delegate",
            "offset_delegate",
            "output_delegate",
            "order_by_delegate",
        ):
            setattr(select, attr, getattr(self, attr))

        if self.prefetch_delegate.fk_columns:
            select.columns(*self.table.all_columns())
            for fk in self.prefetch_delegate.fk_columns:
                if isinstance(fk, ForeignKey):
                    select.columns(*fk.all_columns())
                else:
                    raise ValueError(f"{fk} doesn't seem to be a ForeignKey.")

                # Make sure that all intermediate objects are fully loaded.
                for parent_fk in fk._meta.call_chain:
                    select.columns(*parent_fk.all_columns())

            select.output_delegate.output(nested=True)

        return select.querystrings

    ###########################################################################

    async def run(
        self,
        node: t.Optional[str] = None,
        in_pool: bool = True,
        use_callbacks: bool = True,
    ) -> t.List[TableInstance]:
        results = await super().run(node=node, in_pool=in_pool)

        if use_callbacks:
            # With callbacks, the user can return any data that they want.
            # Assume that most of the time they will still return a list of
            # Table instances.
            modified: t.List[
                TableInstance
            ] = await self.callback_delegate.invoke(
                results, kind=CallbackType.success
            )
            return modified
        else:
            return results

    def __await__(
        self,
    ) -> t.Generator[None, None, t.List[TableInstance]]:
        return super().__await__()


Self = t.TypeVar("Self", bound=Objects)



from piccolo.conf.apps import AppConfig, Command

from .commands.new import new

APP_CONFIG = AppConfig(
    app_name="asgi",
    migrations_folder_path="",
    commands=[Command(callable=new, aliases=["create"])],
)



from __future__ import annotations

import itertools
import typing as t
from dataclasses import dataclass

from piccolo.columns.base import Column
from piccolo.columns.column_types import ForeignKey, Numeric, Varchar
from piccolo.query.base import DDL
from piccolo.utils.warnings import Level, colored_warning

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.columns.base import OnDelete, OnUpdate
    from piccolo.table import Table


class AlterStatement:
    __slots__ = ()  # type: ignore

    @property
    def ddl(self) -> str:
        raise NotImplementedError()

    def __str__(self) -> str:
        return self.ddl


@dataclass
class RenameTable(AlterStatement):
    __slots__ = ("new_name",)

    new_name: str

    @property
    def ddl(self) -> str:
        return f"RENAME TO {self.new_name}"


@dataclass
class AlterColumnStatement(AlterStatement):
    __slots__ = ("column",)

    column: t.Union[Column, str]

    @property
    def column_name(self) -> str:
        if isinstance(self.column, str):
            return self.column
        elif isinstance(self.column, Column):
            return self.column._meta.db_column_name
        else:
            raise ValueError("Unrecognised column type")


@dataclass
class RenameColumn(AlterColumnStatement):
    __slots__ = ("new_name",)

    new_name: str

    @property
    def ddl(self) -> str:
        return f'RENAME COLUMN "{self.column_name}" TO "{self.new_name}"'


@dataclass
class DropColumn(AlterColumnStatement):
    @property
    def ddl(self) -> str:
        return f'DROP COLUMN "{self.column_name}"'


@dataclass
class AddColumn(AlterColumnStatement):
    __slots__ = ("name",)

    column: Column
    name: str

    @property
    def ddl(self) -> str:
        self.column._meta.name = self.name
        return f"ADD COLUMN {self.column.ddl}"


@dataclass
class DropDefault(AlterColumnStatement):
    @property
    def ddl(self) -> str:
        return f'ALTER COLUMN "{self.column_name}" DROP DEFAULT'


@dataclass
class SetColumnType(AlterStatement):
    """
    :param using_expression:
        Postgres can't automatically convert between certain column types. You
        can tell Postgres which action to take. For example
        `my_column_name::integer`.

    """

    old_column: Column
    new_column: Column
    using_expression: t.Optional[str] = None

    @property
    def ddl(self) -> str:
        if self.new_column._meta._table is None:
            self.new_column._meta._table = self.old_column._meta.table

        column_name = self.old_column._meta.db_column_name
        query = (
            f'ALTER COLUMN "{column_name}" TYPE {self.new_column.column_type}'
        )
        if self.using_expression is not None:
            query += f" USING {self.using_expression}"
        return query


@dataclass
class SetDefault(AlterColumnStatement):
    __slots__ = ("value",)

    column: Column
    value: t.Any

    @property
    def ddl(self) -> str:
        sql_value = self.column.get_sql_value(self.value)
        return f'ALTER COLUMN "{self.column_name}" SET DEFAULT {sql_value}'


@dataclass
class SetUnique(AlterColumnStatement):
    __slots__ = ("boolean",)

    boolean: bool

    @property
    def ddl(self) -> str:
        if self.boolean:
            return f'ADD UNIQUE ("{self.column_name}")'
        if isinstance(self.column, str):
            raise ValueError(
                "Removing a unique constraint requires a Column instance "
                "to be passed as the column arg instead of a string."
            )
        tablename = self.column._meta.table._meta.tablename
        column_name = self.column_name
        key = f"{tablename}_{column_name}_key"
        return f'DROP CONSTRAINT "{key}"'


@dataclass
class SetNull(AlterColumnStatement):
    __slots__ = ("boolean",)

    boolean: bool

    @property
    def ddl(self) -> str:
        if self.boolean:
            return f'ALTER COLUMN "{self.column_name}" DROP NOT NULL'
        else:
            return f'ALTER COLUMN "{self.column_name}" SET NOT NULL'


@dataclass
class SetLength(AlterColumnStatement):
    __slots__ = ("length",)

    length: int

    @property
    def ddl(self) -> str:
        return f'ALTER COLUMN "{self.column_name}" TYPE VARCHAR({self.length})'


@dataclass
class DropConstraint(AlterStatement):
    __slots__ = ("constraint_name",)

    constraint_name: str

    @property
    def ddl(self) -> str:
        return f"DROP CONSTRAINT IF EXISTS {self.constraint_name}"


@dataclass
class AddForeignKeyConstraint(AlterStatement):
    __slots__ = (
        "constraint_name",
        "foreign_key_column_name",
        "referenced_table_name",
        "on_delete",
        "on_update",
    )

    constraint_name: str
    foreign_key_column_name: str
    referenced_table_name: str
    on_delete: t.Optional[OnDelete]
    on_update: t.Optional[OnUpdate]
    referenced_column_name: str = "id"

    @property
    def ddl(self) -> str:
        query = (
            f'ADD CONSTRAINT "{self.constraint_name}" FOREIGN KEY '
            f'("{self.foreign_key_column_name}") REFERENCES '
            f'"{self.referenced_table_name}" ("{self.referenced_column_name}")'
        )
        if self.on_delete:
            query += f" ON DELETE {self.on_delete.value}"
        if self.on_update:
            query += f" ON UPDATE {self.on_update.value}"
        return query


@dataclass
class SetDigits(AlterColumnStatement):
    __slots__ = ("digits", "column_type")

    digits: t.Optional[t.Tuple[int, int]]
    column_type: str

    @property
    def ddl(self) -> str:
        if self.digits is None:
            return f'ALTER COLUMN "{self.column_name}" TYPE {self.column_type}'

        precision = self.digits[0]
        scale = self.digits[1]
        return (
            f'ALTER COLUMN "{self.column_name}" TYPE '
            f"{self.column_type}({precision}, {scale})"
        )


@dataclass
class DropTable:
    tablename: str
    cascade: bool
    if_exists: bool

    @property
    def ddl(self) -> str:
        query = "DROP TABLE"

        if self.if_exists:
            query += " IF EXISTS"

        query += f" {self.tablename}"

        if self.cascade:
            query += " CASCADE"

        return query


class Alter(DDL):
    __slots__ = (
        "_add_foreign_key_constraint",
        "_add",
        "_drop_constraint",
        "_drop_default",
        "_drop_table",
        "_drop",
        "_rename_columns",
        "_rename_table",
        "_set_column_type",
        "_set_default",
        "_set_digits",
        "_set_length",
        "_set_null",
        "_set_unique",
    )

    def __init__(self, table: t.Type[Table], **kwargs):
        super().__init__(table, **kwargs)
        self._add_foreign_key_constraint: t.List[AddForeignKeyConstraint] = []
        self._add: t.List[AddColumn] = []
        self._drop_constraint: t.List[DropConstraint] = []
        self._drop_default: t.List[DropDefault] = []
        self._drop_table: t.Optional[DropTable] = None
        self._drop: t.List[DropColumn] = []
        self._rename_columns: t.List[RenameColumn] = []
        self._rename_table: t.List[RenameTable] = []
        self._set_column_type: t.List[SetColumnType] = []
        self._set_default: t.List[SetDefault] = []
        self._set_digits: t.List[SetDigits] = []
        self._set_length: t.List[SetLength] = []
        self._set_null: t.List[SetNull] = []
        self._set_unique: t.List[SetUnique] = []

    def add_column(self: Self, name: str, column: Column) -> Self:
        """
        Band.alter().add_column(‘members’, Integer())
        """
        column._meta._table = self.table
        column._meta._name = name
        column._meta.db_column_name = name

        if isinstance(column, ForeignKey):
            column._setup(table_class=self.table)

        self._add.append(AddColumn(column, name))
        return self

    def drop_column(self, column: t.Union[str, Column]) -> Alter:
        """
        Band.alter().drop_column(Band.popularity)
        """
        self._drop.append(DropColumn(column))
        return self

    def drop_default(self, column: t.Union[str, Column]) -> Alter:
        """
        Band.alter().drop_default(Band.popularity)
        """
        self._drop_default.append(DropDefault(column=column))
        return self

    def drop_table(
        self, cascade: bool = False, if_exists: bool = False
    ) -> Alter:
        """
        Band.alter().drop_table()
        """
        self._drop_table = DropTable(
            tablename=self.table._meta.tablename,
            cascade=cascade,
            if_exists=if_exists,
        )
        return self

    def rename_table(self, new_name: str) -> Alter:
        """
        Band.alter().rename_table('musical_group')
        """
        # We override the existing one rather than appending.
        self._rename_table = [RenameTable(new_name=new_name)]
        return self

    def rename_column(
        self, column: t.Union[str, Column], new_name: str
    ) -> Alter:
        """
        Band.alter().rename_column(Band.popularity, ‘rating’)
        Band.alter().rename_column('popularity', ‘rating’)
        """
        self._rename_columns.append(RenameColumn(column, new_name))
        return self

    def set_column_type(
        self,
        old_column: Column,
        new_column: Column,
        using_expression: t.Optional[str] = None,
    ) -> Alter:
        """
        Change the type of a column.
        """
        self._set_column_type.append(
            SetColumnType(
                old_column=old_column,
                new_column=new_column,
                using_expression=using_expression,
            )
        )
        return self

    def set_default(self, column: Column, value: t.Any) -> Alter:
        """
        Set the default for a column.

        Band.alter().set_default(Band.popularity, 0)
        """
        self._set_default.append(SetDefault(column=column, value=value))
        return self

    def set_null(
        self, column: t.Union[str, Column], boolean: bool = True
    ) -> Alter:
        """
        Band.alter().set_null(Band.name, True)
        Band.alter().set_null('name', True)
        """
        self._set_null.append(SetNull(column, boolean))
        return self

    def set_unique(
        self, column: t.Union[str, Column], boolean: bool = True
    ) -> Alter:
        """
        Band.alter().set_unique(Band.name, True)
        Band.alter().set_unique('name', True)
        """
        self._set_unique.append(SetUnique(column, boolean))
        return self

    def set_length(self, column: t.Union[str, Varchar], length: int) -> Alter:
        """
        Change the max length of a varchar column. Unfortunately, this isn't
        supported by SQLite, but SQLite also doesn't enforce any length limits
        on varchar columns anyway.

        Band.alter().set_length('name', 512)
        """
        if self.engine_type == "sqlite":
            colored_warning(
                (
                    "SQLITE doesn't support changes in length. It also "
                    "doesn't enforce any length limits, so your code will "
                    "still work as expected. Skipping."
                ),
                level=Level.medium,
            )
            return self

        if not isinstance(column, (str, Varchar)):
            raise ValueError(
                "Only Varchar columns can have their length changed."
            )

        self._set_length.append(SetLength(column, length))
        return self

    def _get_constraint_name(self, column: t.Union[str, ForeignKey]) -> str:
        column_name = AlterColumnStatement(column=column).column_name
        tablename = self.table._meta.tablename
        return f"{tablename}_{column_name}_fk"

    def drop_constraint(self, constraint_name: str) -> Alter:
        self._drop_constraint.append(
            DropConstraint(constraint_name=constraint_name)
        )
        return self

    def drop_foreign_key_constraint(
        self, column: t.Union[str, ForeignKey]
    ) -> Alter:
        constraint_name = self._get_constraint_name(column=column)
        return self.drop_constraint(constraint_name=constraint_name)

    def add_foreign_key_constraint(
        self,
        column: t.Union[str, ForeignKey],
        referenced_table_name: str,
        on_delete: t.Optional[OnDelete] = None,
        on_update: t.Optional[OnUpdate] = None,
        referenced_column_name: str = "id",
    ) -> Alter:
        """
        This will add a new foreign key constraint.

        Band.alter().add_foreign_key_constraint(
            Band.manager,
            referenced_table_name='manager',
            on_delete=OnDelete.cascade
        )
        """
        constraint_name = self._get_constraint_name(column=column)
        column_name = AlterColumnStatement(column=column).column_name

        self._add_foreign_key_constraint.append(
            AddForeignKeyConstraint(
                constraint_name=constraint_name,
                foreign_key_column_name=column_name,
                referenced_table_name=referenced_table_name,
                on_delete=on_delete,
                on_update=on_update,
                referenced_column_name=referenced_column_name,
            )
        )
        return self

    def set_digits(
        self,
        column: t.Union[str, Numeric],
        digits: t.Optional[t.Tuple[int, int]],
    ) -> Alter:
        """
        Alter the precision and scale for a Numeric column.
        """
        column_type = (
            column.__class__.__name__.upper()
            if isinstance(column, Numeric)
            else "NUMERIC"
        )
        self._set_digits.append(
            SetDigits(
                digits=digits,
                column=column,
                column_type=column_type,
            )
        )
        return self

    @property
    def default_ddl(self) -> t.Sequence[str]:
        if self._drop_table is not None:
            return [self._drop_table.ddl]

        query = f"ALTER TABLE {self.table._meta.tablename}"

        alterations = [
            i.ddl
            for i in itertools.chain(
                self._add,
                self._rename_columns,
                self._rename_table,
                self._drop,
                self._drop_default,
                self._set_column_type,
                self._set_unique,
                self._set_null,
                self._set_length,
                self._set_default,
                self._set_digits,
            )
        ]

        if self.engine_type == "sqlite":
            # Can only perform one alter statement at a time.
            return [f"{query} {i}" for i in alterations]

        # Postgres can perform them all at once:
        query += ",".join(f" {i}" for i in alterations)

        return [query]


Self = t.TypeVar("Self", bound=Alter)



"""
Most of the tests for piccolo/columns/reference.py are covered in
piccolo/columns/test_foreignkey.py
"""
from unittest import TestCase

from piccolo.columns.reference import LazyTableReference


class TestLazyTableReference(TestCase):
    def test_init(self):
        """
        A ``LazyTableReference`` must be passed either an ``app_name`` or
        ``module_path`` argument.
        """
        with self.assertRaises(ValueError):
            LazyTableReference(table_class_name="Manager")

        with self.assertRaises(ValueError):
            LazyTableReference(
                table_class_name="Manager",
                app_name="music",
                module_path="tests.example_apps.music.tables",
            )

        # Shouldn't raise exceptions:
        LazyTableReference(
            table_class_name="Manager",
            app_name="music",
        )
        LazyTableReference(
            table_class_name="Manager",
            module_path="tests.example_apps.music.tables",
        )

    def test_str(self):
        self.assertEqual(
            LazyTableReference(
                table_class_name="Manager",
                app_name="music",
            ).__str__(),
            "App music.Manager",
        )

        self.assertEqual(
            LazyTableReference(
                table_class_name="Manager",
                module_path="tests.example_apps.music.tables",
            ).__str__(),
            "Module tests.example_apps.music.tables.Manager",
        )



from unittest import TestCase

from tests.example_apps.music.tables import Band


class TestConstructor(TestCase):
    def test_data_parameter(self):
        """
        Make sure the _data parameter works.
        """
        band = Band({Band.name: "Pythonistas"})
        self.assertEqual(band.name, "Pythonistas")

    def test_kwargs(self):
        """
        Make sure kwargs works.
        """
        band = Band(name="Pythonistas")
        self.assertEqual(band.name, "Pythonistas")

    def test_mix(self):
        """
        Make sure the _data paramter and kwargs works together (it's unlikely
        people will do this, but just in case).
        """
        band = Band({Band.name: "Pythonistas"}, popularity=1000)
        self.assertEqual(band.name, "Pythonistas")
        self.assertEqual(band.popularity, 1000)



from __future__ import annotations

import asyncio
import typing as t
from concurrent.futures import ThreadPoolExecutor


def run_sync(coroutine: t.Coroutine):
    """
    Run the coroutine synchronously - trying to accommodate as many edge cases
    as possible.
     1. When called within a coroutine.
     2. When called from ``python -m asyncio``, or iPython with %autoawait
        enabled, which means an event loop may already be running in the
        current thread.
    """
    try:
        # We try this first, as in most situations this will work.
        return asyncio.run(coroutine)
    except RuntimeError:
        # An event loop already exists.
        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(asyncio.run, coroutine)
            return future.result()



import asyncio
import typing as t
from unittest import TestCase

from piccolo.engine.postgres import Atomic
from piccolo.engine.sqlite import SQLiteEngine, TransactionType
from piccolo.table import drop_db_tables_sync
from piccolo.utils.sync import run_sync
from tests.base import engines_only
from tests.example_apps.music.tables import Band, Manager


class TestAtomic(TestCase):
    def test_error(self):
        """
        Make sure queries in a transaction aren't committed if a query fails.
        """
        atomic = Band._meta.db.atomic()
        atomic.add(
            Manager.create_table(),
            Band.create_table(),
            Band.raw("MALFORMED QUERY ... SHOULD ERROR"),
        )
        try:
            atomic.run_sync()
        except Exception:
            pass
        self.assertTrue(not Band.table_exists().run_sync())
        self.assertTrue(not Manager.table_exists().run_sync())

    def test_succeeds(self):
        """
        Make sure that when atomic is run successfully the database is modified
        accordingly.
        """
        atomic = Band._meta.db.atomic()
        atomic.add(Manager.create_table(), Band.create_table())
        atomic.run_sync()

        self.assertTrue(Band.table_exists().run_sync())
        self.assertTrue(Manager.table_exists().run_sync())

        drop_db_tables_sync(Band, Manager)

    @engines_only("postgres", "cockroach")
    def test_pool(self):
        """
        Make sure atomic works correctly when a connection pool is active.
        """

        async def run():
            """
            We have to run this async function, so we can use a connection
            pool.
            """
            engine = Band._meta.db
            await engine.start_connection_pool()

            atomic: Atomic = engine.atomic()
            atomic.add(
                Manager.create_table(),
                Band.create_table(),
            )

            await atomic.run()
            await engine.close_connection_pool()

        run_sync(run())

        self.assertTrue(Band.table_exists().run_sync())
        self.assertTrue(Manager.table_exists().run_sync())

        drop_db_tables_sync(Band, Manager)


class TestTransaction(TestCase):
    def tearDown(self):
        for table in (Band, Manager):
            if table.table_exists().run_sync():
                table.alter().drop_table().run_sync()

    def test_error(self):
        """
        Make sure queries in a transaction aren't committed if a query fails.
        """

        async def run_transaction():
            try:
                async with Band._meta.db.transaction():
                    Manager.create_table()
                    Band.create_table()
                    Band.raw("MALFORMED QUERY ... SHOULD ERROR")
            except Exception:
                pass

        asyncio.run(run_transaction())

        self.assertTrue(not Band.table_exists().run_sync())
        self.assertTrue(not Manager.table_exists().run_sync())

    def test_succeeds(self):
        async def run_transaction():
            async with Band._meta.db.transaction():
                await Manager.create_table().run()
                await Band.create_table().run()

        asyncio.run(run_transaction())

        self.assertTrue(Band.table_exists().run_sync())
        self.assertTrue(Manager.table_exists().run_sync())

    @engines_only("postgres")
    def test_transaction_id(self):
        """
        An extra sanity check, that the transaction id is the same for each
        query inside the transaction block.
        """

        async def run_transaction():
            responses = []
            async with Band._meta.db.transaction():
                responses.append(
                    await Manager.raw("SELECT txid_current()").run()
                )
                responses.append(
                    await Manager.raw("SELECT txid_current()").run()
                )
            return [i[0]["txid_current"] for i in responses]

        txids = asyncio.run(run_transaction())
        self.assertEqual(len(set(txids)), 1)

        # Now run it again and make sure the transaction ids differ.
        next_txids = asyncio.run(run_transaction())
        self.assertNotEqual(txids, next_txids)


class TestTransactionExists(TestCase):
    def test_exists(self):
        """
        Make sure we can detect when code is within a transaction.
        """
        engine = t.cast(SQLiteEngine, Manager._meta.db)

        async def run_inside_transaction():
            async with engine.transaction():
                return engine.transaction_exists()

        self.assertTrue(asyncio.run(run_inside_transaction()))

        async def run_outside_transaction():
            return engine.transaction_exists()

        self.assertFalse(asyncio.run(run_outside_transaction()))


@engines_only("sqlite")
class TestTransactionType(TestCase):
    def setUp(self):
        Manager.create_table().run_sync()

    def tearDown(self):
        Manager.alter().drop_table().run_sync()

    def test_transaction(self):
        """
        With SQLite, we can specify the transaction type. This helps when
        we want to do concurrent writes, to avoid locking the database.

        https://github.com/piccolo-orm/piccolo/issues/687
        """
        engine = t.cast(SQLiteEngine, Manager._meta.db)

        async def run_transaction(name: str):
            async with engine.transaction(
                transaction_type=TransactionType.immediate
            ):
                # This does a SELECT followed by an INSERT, so is a good test.
                # If using TransactionType.deferred it would fail because
                # the database will become locked.
                await Manager.objects().get_or_create(Manager.name == name)

        manager_names = [f"Manager_{i}" for i in range(1, 10)]

        async def run_all():
            """
            Run all of the transactions concurrently.
            """
            await asyncio.gather(
                *[run_transaction(name=name) for name in manager_names]
            )

        asyncio.run(run_all())

        # Make sure it all ran effectively.
        self.assertListEqual(
            Manager.select(Manager.name)
            .order_by(Manager.name)
            .output(as_list=True)
            .run_sync(),
            manager_names,
        )

    def test_atomic(self):
        """
        Similar to above, but with ``Atomic``.
        """
        engine = t.cast(SQLiteEngine, Manager._meta.db)

        async def run_atomic(name: str):
            atomic = engine.atomic(transaction_type=TransactionType.immediate)
            atomic.add(Manager.objects().get_or_create(Manager.name == name))
            await atomic.run()

        manager_names = [f"Manager_{i}" for i in range(1, 10)]

        async def run_all():
            """
            Run all of the transactions concurrently.
            """
            await asyncio.gather(
                *[run_atomic(name=name) for name in manager_names]
            )

        asyncio.run(run_all())

        # Make sure it all ran effectively.
        self.assertListEqual(
            Manager.select(Manager.name)
            .order_by(Manager.name)
            .output(as_list=True)
            .run_sync(),
            manager_names,
        )



import asyncio
from unittest import TestCase

from piccolo.columns.column_types import Varchar
from piccolo.engine.exceptions import TransactionError
from piccolo.engine.sqlite import SQLiteEngine
from piccolo.table import Table
from tests.base import DBTestCase, sqlite_only
from tests.example_apps.music.tables import Manager

ENGINE_1 = SQLiteEngine(path="engine1.sqlite")
ENGINE_2 = SQLiteEngine(path="engine2.sqlite")


class Musician(Table, db=ENGINE_1):
    name = Varchar(length=100)


class Roadie(Table, db=ENGINE_2):
    name = Varchar(length=100)


@sqlite_only
class TestDifferentDB(TestCase):
    def setUp(self):
        ENGINE_1.remove_db_file()
        ENGINE_2.remove_db_file()

    def tearDown(self):
        ENGINE_1.remove_db_file()
        ENGINE_2.remove_db_file()

    async def run_nested(self):
        """
        Make sure nested transactions which reference different databases work
        as expected.
        """
        async with Musician._meta.db.transaction():
            await Musician.create_table().run()
            await Musician(name="Bob").save().run()

            async with Roadie._meta.db.transaction():
                await Roadie.create_table().run()
                await Roadie(name="Dave").save().run()

        self.assertTrue(await Musician.table_exists().run())
        musician = await Musician.select("name").first().run()
        self.assertEqual(musician["name"], "Bob")

        self.assertTrue(await Roadie.table_exists().run())
        roadie = await Roadie.select("name").first().run()
        self.assertEqual(roadie["name"], "Dave")

    def test_nested(self):
        asyncio.run(self.run_nested())


class TestSameDB(DBTestCase):
    async def run_nested(self):
        """
        Nested transactions currently aren't permitted in a connection.
        """
        with self.assertRaises(TransactionError):
            async with Manager._meta.db.transaction():
                await Manager(name="Bob").save().run()

                async with Manager._meta.db.transaction():
                    await Manager(name="Dave").save().run()

    def test_nested(self):
        asyncio.run(self.run_nested())



from __future__ import annotations

import typing as t

from piccolo.columns import Column
from piccolo.columns.indexes import IndexMethod
from piccolo.query.base import DDL

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.table import Table


class CreateIndex(DDL):
    def __init__(
        self,
        table: t.Type[Table],
        columns: t.List[t.Union[Column, str]],
        method: IndexMethod = IndexMethod.btree,
        if_not_exists: bool = False,
        **kwargs,
    ):
        self.columns = columns
        self.method = method
        self.if_not_exists = if_not_exists
        super().__init__(table, **kwargs)

    @property
    def column_names(self) -> t.List[str]:
        return [
            i._meta.db_column_name if isinstance(i, Column) else i
            for i in self.columns
        ]

    @property
    def prefix(self) -> str:
        prefix = "CREATE INDEX"
        if self.if_not_exists:
            prefix += " IF NOT EXISTS"
        return prefix

    @property
    def postgres_ddl(self) -> t.Sequence[str]:
        column_names = self.column_names
        index_name = self.table._get_index_name(column_names)
        tablename = self.table._meta.tablename
        method_name = self.method.value
        column_names_str = ", ".join([f'"{i}"' for i in self.column_names])
        return [
            (
                f"{self.prefix} {index_name} ON {tablename} USING "
                f"{method_name} ({column_names_str})"
            )
        ]

    @property
    def cockroach_ddl(self) -> t.Sequence[str]:
        column_names = self.column_names
        index_name = self.table._get_index_name(column_names)
        tablename = self.table._meta.tablename
        method_name = self.method.value
        column_names_str = ", ".join([f'"{i}"' for i in self.column_names])
        return [
            (
                f"{self.prefix} {index_name} ON {tablename} USING "
                f"{method_name} ({column_names_str})"
            )
        ]

    @property
    def sqlite_ddl(self) -> t.Sequence[str]:
        column_names = self.column_names
        index_name = self.table._get_index_name(column_names)
        tablename = self.table._meta.tablename

        method_name = self.method.value
        if method_name != "btree":
            raise ValueError("SQLite only support btree indexes.")

        column_names_str = ", ".join([f'"{i}"' for i in self.column_names])
        return [
            (
                f"{self.prefix} {index_name} ON {tablename} "
                f"({column_names_str})"
            )
        ]



import inflection


def _camel_to_snake(string: str):
    """
    Convert CamelCase to snake_case.
    """
    return inflection.underscore(string)


def _snake_to_camel(string: str):
    """
    Convert snake_case to CamelCase.
    """
    return inflection.camelize(string)



import datetime
import os
import shutil
import tempfile
from unittest import TestCase
from unittest.mock import MagicMock, call, patch

from piccolo.apps.migrations.commands.new import (
    BaseMigrationManager,
    _create_new_migration,
    _generate_migration_meta,
    new,
)
from piccolo.conf.apps import AppConfig
from piccolo.utils.sync import run_sync
from tests.base import engines_only
from tests.example_apps.music.tables import Manager


class TestNewMigrationCommand(TestCase):
    def test_create_new_migration(self):
        """
        Create a manual migration (i.e. non-auto).
        """
        migration_folder = os.path.join(
            tempfile.gettempdir(), "piccolo_migrations"
        )
        if os.path.exists(migration_folder):
            shutil.rmtree(migration_folder)

        os.mkdir(migration_folder)

        app_config = AppConfig(
            app_name="music",
            migrations_folder_path=migration_folder,
            table_classes=[Manager],
        )

        run_sync(_create_new_migration(app_config, auto=False))

        migration_modules = BaseMigrationManager().get_migration_modules(
            migration_folder
        )

        self.assertTrue(len(migration_modules.keys()) == 1)

    @engines_only("postgres")
    @patch("piccolo.apps.migrations.commands.new.print")
    def test_new_command(self, print_: MagicMock):
        """
        Call the command, when no migration changes are needed.
        """
        with self.assertRaises(SystemExit) as manager:
            run_sync(new(app_name="music", auto=True))

        self.assertEqual(manager.exception.code, 0)

        self.assertTrue(
            print_.mock_calls[-1] == call("No changes detected - exiting.")
        )


class TestGenerateMigrationMeta(TestCase):
    @patch("piccolo.apps.migrations.commands.new.now")
    def test_filename(self, now: MagicMock):
        now.return_value = datetime.datetime(
            year=2022,
            month=1,
            day=10,
            hour=7,
            minute=15,
            second=20,
            microsecond=3000,
        )

        # Try with an app name which already contains valid characters for a
        # Python module.
        migration_meta = _generate_migration_meta(
            app_config=AppConfig(
                app_name="app_name",
                migrations_folder_path="/tmp/",
            )
        )
        self.assertEqual(
            migration_meta.migration_filename,
            "app_name_2022_01_10t07_15_20_003000",
        )
        self.assertEqual(
            migration_meta.migration_path,
            "/tmp/app_name_2022_01_10t07_15_20_003000.py",
        )

        # Try with an app name with invalid characters for a Python module.
        migration_meta = _generate_migration_meta(
            app_config=AppConfig(
                app_name="App-Name!",
                migrations_folder_path="/tmp/",
            )
        )
        self.assertEqual(
            migration_meta.migration_filename,
            "app_name_2022_01_10t07_15_20_003000",
        )
        self.assertEqual(
            migration_meta.migration_path,
            "/tmp/app_name_2022_01_10t07_15_20_003000.py",
        )



from __future__ import annotations

import typing as t

from piccolo.custom_types import Combinable
from piccolo.query.base import Query
from piccolo.query.mixins import ReturningDelegate, WhereDelegate
from piccolo.querystring import QueryString

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.columns import Column
    from piccolo.table import Table


class DeletionError(Exception):
    pass


class Delete(Query):

    __slots__ = (
        "force",
        "returning_delegate",
        "where_delegate",
    )

    def __init__(self, table: t.Type[Table], force: bool = False, **kwargs):
        super().__init__(table, **kwargs)
        self.force = force
        self.returning_delegate = ReturningDelegate()
        self.where_delegate = WhereDelegate()

    def where(self: Self, *where: Combinable) -> Self:
        self.where_delegate.where(*where)
        return self

    def returning(self: Self, *columns: Column) -> Self:
        self.returning_delegate.returning(columns)
        return self

    def _validate(self):
        """
        Don't let a deletion happen unless it has a where clause, or is
        explicitly forced.
        """
        if (not self.where_delegate._where) and (not self.force):
            classname = self.table.__name__
            raise DeletionError(
                "Do you really want to delete all the data from "
                f"{classname}? If so, use {classname}.delete(force=True). "
                "Otherwise, add a where clause."
            )

    @property
    def default_querystrings(self) -> t.Sequence[QueryString]:
        query = f"DELETE FROM {self.table._meta.tablename}"

        querystring = QueryString(query)

        if self.where_delegate._where:
            querystring = QueryString(
                "{} WHERE {}",
                querystring,
                self.where_delegate._where.querystring,
            )

        if self.returning_delegate._returning:
            querystring = QueryString(
                "{}{}",
                querystring,
                self.returning_delegate._returning.querystring,
            )

        return [querystring]


Self = t.TypeVar("Self", bound=Delete)



from decimal import Decimal

from piccolo.apps.migrations.auto import MigrationManager
from piccolo.columns.base import OnDelete, OnUpdate
from piccolo.columns.column_types import (
    JSON,
    JSONB,
    UUID,
    BigInt,
    Boolean,
    Bytea,
    Date,
    DoublePrecision,
    ForeignKey,
    Integer,
    Interval,
    Numeric,
    Real,
    Serial,
    SmallInt,
    Text,
    Timestamp,
    Timestamptz,
    Varchar,
)
from piccolo.columns.defaults.date import DateNow
from piccolo.columns.defaults.interval import IntervalCustom
from piccolo.columns.defaults.timestamp import TimestampNow
from piccolo.columns.defaults.timestamptz import TimestamptzNow
from piccolo.columns.defaults.uuid import UUID4
from piccolo.columns.indexes import IndexMethod
from piccolo.table import Table


class SmallTable(Table, tablename="small_table"):
    id = Serial(
        null=False,
        primary_key=True,
        unique=False,
        index=False,
        index_method=IndexMethod.btree,
        choices=None,
    )


ID = "2021-09-20T21:23:25:698988"
VERSION = "0.49.0"
DESCRIPTION = ""


async def forwards():
    manager = MigrationManager(
        migration_id=ID, app_name="mega", description=DESCRIPTION
    )

    manager.add_table("MegaTable", tablename="mega_table")

    manager.add_table("SmallTable", tablename="small_table")

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="bigint_col",
        column_class_name="BigInt",
        column_class=BigInt,
        params={
            "default": 0,
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="boolean_col",
        column_class_name="Boolean",
        column_class=Boolean,
        params={
            "default": False,
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="bytea_col",
        column_class_name="Bytea",
        column_class=Bytea,
        params={
            "default": b"",
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="date_col",
        column_class_name="Date",
        column_class=Date,
        params={
            "default": DateNow(),
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="foreignkey_col",
        column_class_name="ForeignKey",
        column_class=ForeignKey,
        params={
            "references": SmallTable,
            "on_delete": OnDelete.cascade,
            "on_update": OnUpdate.cascade,
            "null": True,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="integer_col",
        column_class_name="Integer",
        column_class=Integer,
        params={
            "default": 0,
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="interval_col",
        column_class_name="Interval",
        column_class=Interval,
        params={
            "default": IntervalCustom(
                weeks=0,
                days=0,
                hours=0,
                minutes=0,
                seconds=0,
                milliseconds=0,
                microseconds=0,
            ),
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="json_col",
        column_class_name="JSON",
        column_class=JSON,
        params={
            "default": "{}",
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="jsonb_col",
        column_class_name="JSONB",
        column_class=JSONB,
        params={
            "default": "{}",
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="numeric_col",
        column_class_name="Numeric",
        column_class=Numeric,
        params={
            "default": Decimal("0"),
            "digits": (5, 2),
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="real_col",
        column_class_name="Real",
        column_class=Real,
        params={
            "default": 0.0,
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="double_precision_col",
        column_class_name="DoublePrecision",
        column_class=DoublePrecision,
        params={
            "default": 0.0,
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="smallint_col",
        column_class_name="SmallInt",
        column_class=SmallInt,
        params={
            "default": 0,
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="text_col",
        column_class_name="Text",
        column_class=Text,
        params={
            "default": "",
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="timestamp_col",
        column_class_name="Timestamp",
        column_class=Timestamp,
        params={
            "default": TimestampNow(),
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="timestamptz_col",
        column_class_name="Timestamptz",
        column_class=Timestamptz,
        params={
            "default": TimestamptzNow(),
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="uuid_col",
        column_class_name="UUID",
        column_class=UUID,
        params={
            "default": UUID4(),
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="varchar_col",
        column_class_name="Varchar",
        column_class=Varchar,
        params={
            "length": 255,
            "default": "",
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="unique_col",
        column_class_name="Varchar",
        column_class=Varchar,
        params={
            "length": 255,
            "default": "",
            "null": False,
            "primary_key": False,
            "unique": True,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="null_col",
        column_class_name="Varchar",
        column_class=Varchar,
        params={
            "length": 255,
            "default": "",
            "null": True,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="MegaTable",
        tablename="mega_table",
        column_name="not_null_col",
        column_class_name="Varchar",
        column_class=Varchar,
        params={
            "length": 255,
            "default": "",
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    manager.add_column(
        table_class_name="SmallTable",
        tablename="small_table",
        column_name="varchar_col",
        column_class_name="Varchar",
        column_class=Varchar,
        params={
            "length": 255,
            "default": "",
            "null": False,
            "primary_key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
            "choices": None,
        },
    )

    return manager



from unittest import TestCase

from tests.example_apps.music.tables import Manager


class TestTableExists(TestCase):
    def setUp(self):
        Manager.create_table().run_sync()

    def test_table_exists(self):
        response = Manager.table_exists().run_sync()
        self.assertTrue(response)

    def tearDown(self):
        Manager.alter().drop_table().run_sync()



"""
Populates a database with an example schema and data, and launches a shell
for interacting with the data using Piccolo.
"""
import datetime
import sys
import typing as t
import uuid
from decimal import Decimal
from enum import Enum

from piccolo.columns import (
    JSON,
    UUID,
    Boolean,
    ForeignKey,
    Integer,
    Interval,
    Numeric,
    Timestamp,
    Varchar,
)
from piccolo.columns.readable import Readable
from piccolo.engine import PostgresEngine, SQLiteEngine
from piccolo.engine.base import Engine
from piccolo.table import Table
from piccolo.utils.warnings import colored_string


class Manager(Table):
    name = Varchar(length=50)

    @classmethod
    def get_readable(cls) -> Readable:
        return Readable(
            template="%s",
            columns=[cls.name],
        )


class Band(Table):
    name = Varchar(length=50)
    manager = ForeignKey(references=Manager, null=True)
    popularity = Integer()

    @classmethod
    def get_readable(cls) -> Readable:
        return Readable(
            template="%s",
            columns=[cls.name],
        )


class Venue(Table):
    name = Varchar(length=100)
    capacity = Integer(default=0)

    @classmethod
    def get_readable(cls) -> Readable:
        return Readable(
            template="%s",
            columns=[cls.name],
        )


class Concert(Table):
    band_1 = ForeignKey(Band)
    band_2 = ForeignKey(Band)
    venue = ForeignKey(Venue)
    starts = Timestamp()
    duration = Interval()

    @classmethod
    def get_readable(cls) -> Readable:
        return Readable(
            template="%s and %s at %s",
            columns=[
                cls.band_1.name,
                cls.band_2.name,
                cls.venue.name,
            ],
        )


class Ticket(Table):
    class TicketType(Enum):
        sitting = "sitting"
        standing = "standing"
        premium = "premium"

    concert = ForeignKey(Concert)
    price = Numeric(digits=(5, 2))
    ticket_type = Varchar(choices=TicketType, default=TicketType.standing)

    @classmethod
    def get_readable(cls) -> Readable:
        return Readable(
            template="%s - %s",
            columns=[
                t.cast(t.Type[Venue], cls.concert.venue).name,
                cls.ticket_type,
            ],
        )


class DiscountCode(Table):
    code = UUID()
    active = Boolean(default=True, null=True)

    @classmethod
    def get_readable(cls) -> Readable:
        return Readable(
            template="%s - %s",
            columns=[cls.code, cls.active],
        )


class RecordingStudio(Table):
    name = Varchar(length=100)
    facilities = JSON(null=True)

    @classmethod
    def get_readable(cls) -> Readable:
        return Readable(
            template="%s",
            columns=[cls.name],
        )


TABLES = (Manager, Band, Venue, Concert, Ticket, DiscountCode, RecordingStudio)


def populate():
    """
    Drop then recreate the tables, and populate with data.
    """
    for _table in reversed(TABLES):
        try:
            if _table.table_exists().run_sync():
                _table.alter().drop_table().run_sync()
        except Exception as e:
            print(e)

    for _table in TABLES:
        try:
            _table.create_table().run_sync()
        except Exception as e:
            print(e)

    guido = Manager(name="Guido")
    guido.save().run_sync()

    pythonistas = Band(name="Pythonistas", manager=guido.id, popularity=1000)
    pythonistas.save().run_sync()

    graydon = Manager(name="Graydon")
    graydon.save().run_sync()

    rustaceans = Band(name="Rustaceans", manager=graydon.id, popularity=500)
    rustaceans.save().run_sync()

    anders = Manager(name="Anders")
    anders.save().run_sync()

    c_sharps = Band(name="C-Sharps", popularity=700, manager=anders.id)
    c_sharps.save().run_sync()

    venue = Venue(name="Amazing Venue", capacity=5000)
    venue.save().run_sync()

    concert = Concert(
        band_1=pythonistas.id,
        band_2=rustaceans.id,
        venue=venue.id,
        starts=datetime.datetime.now() + datetime.timedelta(days=7),
        duration=datetime.timedelta(hours=2),
    )
    concert.save().run_sync()

    ticket = Ticket(concert=concert.id, price=Decimal("50.0"))
    ticket.save().run_sync()

    discount_code = DiscountCode(code=uuid.uuid4())
    discount_code.save().run_sync()

    recording_studio = RecordingStudio(
        name="Abbey Road", facilities={"restaurant": True, "mixing_desk": True}
    )
    recording_studio.save().run_sync()


def run(
    engine: str = "sqlite",
    user: str = "piccolo",
    password: str = "piccolo",
    database: str = "piccolo_playground",
    host: str = "localhost",
    port: int = 5432,
    ipython_profile: bool = False,
):
    """
    Creates a test database to play with.

    :param engine:
        Which database engine to use - options are sqlite or postgres
    :param user:
        Postgres user
    :param password:
        Postgres password
    :param database:
        Postgres database
    :param host:
        Postgres host
    :param port:
        Postgres port
    :param ipython_profile:
        Set to true to use your own IPython profile. Located at ~/.ipython/.
        For more info see the IPython docs
        https://ipython.readthedocs.io/en/stable/config/intro.html.
    """
    try:
        import IPython
    except ImportError:
        sys.exit(
            "Install iPython using `pip install 'piccolo[playground,sqlite]'` "
            "to use this feature."
        )

    if engine.upper() == "POSTGRES":
        db: Engine = PostgresEngine(
            {
                "host": host,
                "database": database,
                "user": user,
                "password": password,
                "port": port,
            }
        )
    else:
        db = SQLiteEngine()
    for _table in TABLES:
        _table._meta._db = db

    print(colored_string("\nTables:\n"))

    for _table in TABLES:
        print(_table._table_str(abbreviated=True))
        print("")

    print(colored_string("Try it as a query builder:"))
    print("await Band.select()")
    print("await Band.select(Band.name)")
    print("await Band.select(Band.name, Band.manager.name)")
    print("\n")

    print(colored_string("Try it as an ORM:"))
    print("b = await Band.objects().where(Band.name == 'Pythonistas').first()")
    print("b.popularity = 10000")
    print("await b.save()")
    print("\n")

    populate()

    from IPython.core.interactiveshell import _asyncio_runner

    if ipython_profile:
        print(colored_string("Using your IPython profile\n"))
        # To try this out, set `c.TerminalInteractiveShell.colors = "Linux"`
        # in `~/.ipython/profile_default/ipython_config.py` to set the terminal
        # color.
        conf_args = {}
    else:
        conf_args = {"colors": "neutral"}

    IPython.embed(using=_asyncio_runner, **conf_args)
